/************************************************************************/
/* der.c - DER encoding/decoding Source                                 */
/************************************************************************/
/*
 *  (c)2007 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <cutil/cutil_portable.h>
#include <cutil/der.h>


/* internal function prototypes */
int der_identifier_encode_r (DERIdentifier *di,
				char *data,
				int *datalenp);
int der_identifier_child_append_decode (DERIdentifier *parent,
				DERIdentifier *di);


CUTIL_API int der_identifier_new (DERIdentifier **dip)
{
	DERIdentifier	*di = NULL;
	if (dip == NULL)
		return EINVAL;
	*dip = NULL;
	di = calloc (1, sizeof (DERIdentifier));
	if (di == NULL)
		return ENOMEM;
	*dip = di;
	return 0;
}


CUTIL_API int der_identifier_new_boolean (DERIdentifier **dip, int value)
{
	int		rc;
	DERIdentifier	*di = NULL;
	if (dip == NULL)
		return EINVAL;
	*dip = NULL;
	rc = der_identifier_new (&di);
	if (rc != 0)
		return rc;
	di->identifier = DER_TYPE_BOOLEAN;
	if (value != 0)
		di->l = 1;
	else
		di->l = 0;
	di->length = 1;
	*dip = di;
	return 0;
}


CUTIL_API int der_identifier_new_integer (DERIdentifier **dip, int value)
{
	int		rc;
	DERIdentifier	*di = NULL;
	if (dip == NULL)
		return EINVAL;
	*dip = NULL;
	rc = der_identifier_new (&di);
	if (rc != 0)
		return rc;
	di->identifier = DER_TYPE_INTEGER;
	di->l = (long)value;
	if (value > -127 && value <= 127)
		di->length = 1;
	else if (value > -32767 && value <= 32767)
		di->length = 2;
	else
		di->length = 4;
	*dip = di;
	return 0;
}


CUTIL_API int der_identifier_new_string (DERIdentifier **dip, char *string, int length)
{
	int		rc;
	DERIdentifier	*di = NULL;
	if (dip == NULL)
		return EINVAL;
	*dip = NULL;
	rc = der_identifier_new (&di);
	if (rc != 0)
		return rc;
	di->identifier = DER_TYPE_OCTETSTRING;
	di->s = calloc (1, length);
	if (di->s == NULL) {
		rc = der_identifier_free (&di);
		return ENOMEM;
	}
	memcpy (di->s, string, length);
	di->length = length;
	*dip = di;
	return 0;
}


CUTIL_API int der_identifier_new_null (DERIdentifier **dip)
{
	int		rc;
	DERIdentifier	*di = NULL;
	if (dip == NULL)
		return EINVAL;
	*dip = NULL;
	rc = der_identifier_new (&di);
	if (rc != 0)
		return rc;
	di->identifier = DER_TYPE_NULL;
	*dip = di;
	return 0;
}


CUTIL_API int der_identifier_new_oid (DERIdentifier **dip, char *string, int length)
{
	return EINVAL;
}


CUTIL_API int der_identifier_new_sequence (DERIdentifier **dip)
{
	int		rc;
	DERIdentifier	*di = NULL;
	if (dip == NULL)
		return EINVAL;
	*dip = NULL;
	rc = der_identifier_new (&di);
	if (rc != 0)
		return rc;
	di->identifier = DER_TYPE_SEQUENCE;
	*dip = di;
	return 0;
}


CUTIL_API int der_identifier_free (DERIdentifier **dip)
{
	int		rc;
	DERIdentifier	*di = NULL;
	if (dip == NULL || *dip == NULL)
		return EINVAL;
	di = *dip;
	switch (di->identifier) {
		case DER_TYPE_BITSTRING:
		case DER_TYPE_OCTETSTRING:
		case DER_TYPE_OID:
			if (di->s != NULL)
				free (di->s);
			break;
		case DER_TYPE_SEQUENCE:
			rc = der_identifier_chain_free (&di->child);
			break;
		default:
			break;
	}
	free (di);
	*dip = NULL;
	return 0;
}


CUTIL_API int der_identifier_get_boolean (DERIdentifier *di, int *valuep)
{
	if (di == NULL || valuep == NULL)
		return EINVAL;
	if (di->identifier != DER_TYPE_BOOLEAN)
		return EINVAL;
	*valuep = di->l;
	return 0;
}


CUTIL_API int der_identifier_get_integer (DERIdentifier *di, int *valuep)
{
	if (di == NULL || valuep == NULL)
		return EINVAL;
	if (di->identifier != DER_TYPE_INTEGER)
		return EINVAL;
	*valuep = di->l;
	return 0;
}


CUTIL_API int der_identifier_get_octetstring (DERIdentifier *di, char *str, int *lengthp)
{
	if (di == NULL || str == NULL || lengthp == NULL)
		return EINVAL;
	if (di->identifier != DER_TYPE_OCTETSTRING)
		return EINVAL;
	if (di->length > *lengthp)
		return ENOSPC;
	memcpy (str, di->s, di->length);
	*lengthp = di->length;
	return 0;
}


CUTIL_API int der_identifier_get_oid (DERIdentifier *di, char *str, int *lengthp)
{
	if (di == NULL || str == NULL || lengthp == NULL)
		return EINVAL;
	if (di->identifier != DER_TYPE_OID)
		return EINVAL;
	return 0;
}


CUTIL_API int der_identifier_chain_append (DERIdentifier *parent, DERIdentifier *di)
{
	if (parent == NULL || di == NULL)
		return EINVAL;
	while (parent->next != NULL)
		parent = parent->next;
	parent->next = di;
	return 0;
}


CUTIL_API int der_identifier_chain_prepend (DERIdentifier *parent, DERIdentifier *di)
{
	if (parent == NULL || di == NULL)
		return EINVAL;
	while (parent->previous != NULL)
		parent = parent->previous;
	parent->previous = di;
	return 0;
}


CUTIL_API int der_identifier_chain_free (DERIdentifier **dip)
{
	int		rc;
	DERIdentifier	*di = NULL;
	DERIdentifier	*item = NULL;
	if (dip == NULL || *dip == NULL)
		return EINVAL;
	item = *dip;
	if (item->previous != NULL) {
		di = item->previous;
		di->next = NULL;
	}
	while (item != NULL) {
		di = item->next;
		rc = der_identifier_free (&item);
		item = di;
	}
	return 0;
}


CUTIL_API int der_identifier_child_append (DERIdentifier *parent, DERIdentifier *di)
{
	DERIdentifier	*item = NULL;
	if (parent == NULL || di == NULL)
		return EINVAL;
	if (parent->identifier != DER_TYPE_SEQUENCE)
		return EINVAL;
	if (parent->child == NULL) {
		parent->child = di;
	} else {
		item = parent->child;
		while (item->next != NULL)
			item = item->next;
		item->next = di;
	}
	parent->length++;
	if (di->length < 128)
		parent->length += 1;
	else if (di->length < 65535)
		parent->length += 2;
	else
		parent->length += 4;
	parent->length += di->length;
	return 0;
}


CUTIL_API int der_identifier_child_prepend (DERIdentifier *parent, DERIdentifier *di)
{
	DERIdentifier	*item = NULL;
	if (parent == NULL || di == NULL)
		return EINVAL;
	if (parent->identifier != DER_TYPE_SEQUENCE)
		return EINVAL;
	if (parent->child == NULL) {
		parent->child = di;
	} else {
		item = parent->child;
		item->previous = di;
		di->next = item;
		parent->child = di;
	}
	parent->length++;
	if (di->length < 128)
		parent->length += 1;
	else if (di->length < 65535)
		parent->length += 2;
	else
		parent->length += 4;
	parent->length += di->length;
	return 0;
}


CUTIL_API int der_identifier_encode (DERIdentifier *di, char **datap, int *datalenp)
{
	int		rc;
	int		len;
	int		enclen;
	char		*p = NULL;
	char		*derdata = NULL;
	DERIdentifier	*item = NULL;
	if (di == NULL || datap == NULL || datalenp == NULL)
		return EINVAL;
	*datap = NULL;
	*datalenp = 0;

	enclen = 0;
	item = di;
	while (item != NULL) {
		enclen++;
		if (item->length < 128)
			enclen += 1;
		else if (item->length < 65535)
			enclen += 2;
		else
			enclen += 4;
		enclen += item->length;
		item = item->next;
	}
	derdata = calloc (1, enclen);
	if (derdata == NULL)
		return ENOMEM;
	p = derdata;
	item = di;
	while (item != NULL) {
		len = enclen - (p - derdata);
		rc = der_identifier_encode_r (item, p, &len);
		if (rc != 0) {
			free (derdata);
			return rc;
		}
		p += len;
		item = item->next;
	}
	*datap = derdata;
	*datalenp = enclen;
	return 0;
}


CUTIL_API int der_identifier_decode (DERIdentifier **dip, char *data, int datalen)
{
	int		rc;
	long		len;
	char		*p = NULL;
	char		*pend = NULL;
	DERIdentifier	*dib = NULL;
	DERIdentifier	*dic = NULL;
	DERIdentifier	*di = NULL;
	if (dip == NULL)
		return EINVAL;
	*dip = NULL;
	if (data == NULL || datalen < 2)
		return EINVAL;
	p = data;
	pend = p + datalen;
	while (p < pend) {
		rc = der_identifier_new (&di);
		if (rc != 0) {
			rc = der_identifier_chain_free (&dib);
			return ENOMEM;
		}
		/* identifier */
		switch (*p) {
			case DER_TYPE_BOOLEAN:
				break;
			case DER_TYPE_INTEGER:
				break;
			case DER_TYPE_OCTETSTRING:
				break;
			case DER_TYPE_NULL:
				break;
			case DER_TYPE_SEQUENCE:
				break;
			default:
				rc = der_identifier_free (&di);
				rc = der_identifier_chain_free (&dib);
				return EINVAL;
		}
		di->identifier = *p;
		p++;
		if (p >= pend) {
			rc = der_identifier_free (&di);
			rc = der_identifier_chain_free (&dib);
			return EINVAL;
		}
		/* length */
		if (*p & 0x80) {
			len = *p & 0x7f;
			p++;
			if (len == 2) {
				len = *(short *)p;
				p += 2;
			} else if (len == 4) {
				len = *(long *)p;
				p += 4;
			} else {
				rc = der_identifier_free (&di);
				rc = der_identifier_chain_free (&dib);
				return EINVAL;
			}
		} else {
			len = (int)*p;
			p++;
		}
		di->length = len;

		if (p > pend || (p+len) > pend) {
			rc = der_identifier_free (&di);
			rc = der_identifier_chain_free (&dib);
			return EINVAL;
		}
		/* value */
		switch (di->identifier) {
			case DER_TYPE_BOOLEAN:
				di->l = *p;
				break;
			case DER_TYPE_INTEGER:
				switch (di->length) {
					case 1:
						di->l = *p;
						break;
					case 2:
						di->l = ntohs (*(short *)p);
						break;
					case 4:
						di->l = ntohl (*(long *)p);
						break;
				}
				break;
			case DER_TYPE_OCTETSTRING:
				di->s = calloc (1, di->length+1);
				if (di->s == NULL) {
					rc = der_identifier_free (&di);
					rc = der_identifier_chain_free (&dib);
					return ENOMEM;
				}
				memcpy (di->s, p, di->length);
				break;
			case DER_TYPE_NULL:
				break;
			case DER_TYPE_SEQUENCE:
				rc = der_identifier_decode (&dic,
							p,
							di->length);
				if (rc != 0) {
					rc = der_identifier_free (&di);
					rc = der_identifier_chain_free (&dib);
					return rc;
				}
				rc = der_identifier_child_append_decode (di,
									dic);
				if (rc != 0) {
					rc = der_identifier_chain_free (&dic);
					rc = der_identifier_free (&di);
					rc = der_identifier_chain_free (&dib);
					return rc;
				}
				break;
		}
		p += di->length;
		/* chain */
		if (dib == NULL)
			dib = di;
		else
			rc = der_identifier_chain_append (dib, di);
	}
	*dip = dib;
	return 0;
}


int der_identifier_encode_r (DERIdentifier *di,
				char *data,
				int *datalenp)
{
	int		rc;
	int		len;
	int		enclen;
	char		*p = NULL;
	DERIdentifier	*item = NULL;
	if (di == NULL || data == NULL || datalenp == NULL)
		return EINVAL;
	enclen = 1;
	if (di->length < 128)
		enclen += 1;
	else if (di->length < 65535)
		enclen += 2;
	else
		enclen += 4;
	enclen += di->length;
	if (*datalenp < enclen)
		return ENOSPC;
	p = data;

	*p = di->identifier;
	p++;

	if (di->length < 128) {
		*p = di->length;
	} else if (di->length < 65536) {
		*p = 0x80 + 0x02;
		p++;
		*p = (di->length & 0x0000ff00) >> 8;
		p++;
		*p = di->length & 0x000000ff;
	} else {
		*p = 0x80 + 0x04;
		p++;
		*p = (di->length & 0xff000000) >> 24;
		p++;
		*p = (di->length & 0x00ff0000) >> 16;
		p++;
		*p = (di->length & 0x0000ff00) >> 8;
		p++;
		*p = di->length & 0x000000ff;
	}
	p++;

	switch (di->identifier) {
		case DER_TYPE_BOOLEAN:
			*p = (char)di->l;
			p++;
			break;
		case DER_TYPE_INTEGER:
			switch (di->length) {
				case 1:
					*p = (char)di->l;
					p++;
					break;
				case 2:
					*(short *)p = htons ((short)di->l);
					p += 2;
					break;
				case 4:
					*(long *)p = htonl (di->l);
					p += 4;
					break;
				default:
					return EINVAL;
			}
			break;
		case DER_TYPE_OCTETSTRING:
			memcpy (p, di->s, di->length);
			p += di->length;
			break;
		case DER_TYPE_NULL:
		/*	*p = 0;
			p++; */
			break;
		case DER_TYPE_SEQUENCE:
			item = di->child;
			while (item != NULL) {
				len = *datalenp - (p - data);
				rc = der_identifier_encode_r (item,
							p,
							&len);
				if (rc != 0)
					return rc;
				p += len;
				item = item->next;
			}
			break;
		default:
			return EINVAL;
	}
	*datalenp = enclen;
	return 0;
}


/* The function is called from der_identifier_decode when the child(ren)
   is/are added to the SEQUENCE parent. This function does not increment
   the parent's ->length member, because that is already known during
   decoding.
*/
int der_identifier_child_append_decode (DERIdentifier *parent,
				DERIdentifier *di)
{
	DERIdentifier	*item = NULL;
	if (parent == NULL || di == NULL)
		return EINVAL;
	if (parent->identifier != DER_TYPE_SEQUENCE)
		return EINVAL;
	if (parent->child == NULL) {
		parent->child = di;
	} else {
		item = parent->child;
		while (item->next != NULL)
			item = item->next;
		item->next = di;
	}
	return 0;
}


