/************************************************************************
 *  ser.c - Simple Encoding Rules source
 ************************************************************************
 *  (c)2007 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

#include <cutil/cutil_portable.h>
#include <cutil/ser.h>

/* 64bit hton defines */
#define	hton64(i)	( ((uint64_t)(htonl((i) & 0xffffffff)) << 32) | htonl(((i) >> 32) & 0xffffffff ) )
#define	ntoh64		hton64

#define	ELEMENT_TYPE(x)		(x & 0x7ffc0000)
#define ELEMENT_PAD(x)		((x & 0x00030000) >> 16)
#define ELEMENT_LENGTH(x)	(x & 0x0000ffff)
#define ELEMENT_IS_SIGNED(x)	(x & SER_TYPE_SIGNED)
#define ELEMENT_SHIFT_PAD(x)	((x) << 16)


#define SER_PACKET_FLAG_ARRAYLOCKED	1


struct	_SERElement {
	uint32_t	refs; 		/* reference counting */
	SERElement	*next;		/* next element pointer */
	SERElement	*prev;		/* previous element */
	uint32_t	type;		/* bitfield: */
					/* 0-15 = int16 value */
					/* 16-17 = # of pad bytes */
					/* 18-32 = element type */
	union {
		int32_t int32;
		uint32_t uint32;
		int64_t int64;
		uint64_t uint64;
		float sfloat;
		char *octets;
		void *svoid;
	} val;				/* data storage union */
};

struct	_SERPacket {
	uint32_t	refs;		/* reference counting */
	SERElement	*cache;		/* pre-allocated elements */
	SERElement	*head;		/* head element */
	SERElement	*tail;		/* tail element */
	SERElement	*iter;		/* iterator */
	uint32_t	elements;	/* # elements in packet */
	uint32_t	length;		/* # bytes in packet */
	uint32_t	flags;		/* packet flags */
};


CUTIL_API int ser_element_new (SERElement **sep, int type, void *valuep, int length)
{
	int		rc = 0;
	SERElement	*se = NULL;
	if (sep == NULL || valuep == NULL)
		return EINVAL;
	*sep = NULL;
	switch (ELEMENT_TYPE (type)) {
		case SER_TYPE_INTEGER16:
			if (ELEMENT_IS_SIGNED (type))
				rc = ser_element_new_int16 (&se,
						*(int16_t *)valuep);
			else
				rc = ser_element_new_uint16 (&se,
						*(uint16_t *)valuep);
			break;
		case SER_TYPE_INTEGER32:
			if (ELEMENT_IS_SIGNED (type))
				rc = ser_element_new_int32 (&se,
						*(int32_t *)valuep);
			else
				rc = ser_element_new_uint32 (&se,
						*(uint32_t *)valuep);
			break;
		case SER_TYPE_INTEGER64:
			if (ELEMENT_IS_SIGNED (type))
				rc = ser_element_new_int64 (&se,
						*(int64_t *)valuep);
			else
				rc = ser_element_new_uint64 (&se,
						*(uint64_t *)valuep);
			break;
		case SER_TYPE_FLOAT:
			rc = ser_element_new_float (&se,
						*(float *)valuep);
			break;
		case SER_TYPE_STRING:
			rc = ser_element_new_string_length (&se,
						(char *)valuep,
						length);
			break;
		case SER_TYPE_BINARY:
			rc = ser_element_new_binary (&se,
						(char *)valuep,
						length);
			break;
/*
		case SER_TYPE_ARRAY:
			rc = ser_element_new_array (&se);
			break;
*/
		default:
			return EINVAL;
	}
	*sep = se;
	return rc;
}


CUTIL_API int ser_element_free (SERElement **sep)
{
	int		rc;
	SERElement	*se = NULL;
	if (sep == NULL || *sep == NULL)
		return EINVAL;
	se = *sep;
	if (se->type & (SER_TYPE_STRING|SER_TYPE_BINARY)) {
		free (se->val.octets);
	} else if (se->type & SER_TYPE_ARRAY) {
		rc = ser_packet_free ((SERPacket **)&se->val.svoid);
		if (rc != 0)
			return rc;
	}
	free (se);
	*sep = NULL;
	return 0;
}


CUTIL_API int ser_element_new_uint16 (SERElement **sep, uint16_t value)
{
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_INTEGER16;
	se->type |= value;
	*sep = se;
	return 0;
}


CUTIL_API int ser_element_new_int16 (SERElement **sep, int16_t value)
{
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_INTEGER16 | SER_TYPE_SIGNED;
	// cast to unsigned before adding to se->type to prevent
	// the MSB from being converted to a 32bit value
	se->type |= (uint16_t)value;
	*sep = se;
	return 0;
}


CUTIL_API int ser_element_new_uint32 (SERElement **sep, uint32_t value)
{
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_INTEGER32;
	se->val.uint32 = value;
	se->type |= sizeof (uint32_t);
	*sep = se;
	return 0;
}


CUTIL_API int ser_element_new_int32 (SERElement **sep, int32_t value)
{
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_INTEGER32 | SER_TYPE_SIGNED;
	se->val.int32 = value;
	se->type |= sizeof (int32_t);
	*sep = se;
	return 0;
}


CUTIL_API int ser_element_new_uint64 (SERElement **sep, uint64_t value)
{
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_INTEGER64;
	se->val.uint64 = value;
	se->type |= sizeof (uint64_t);
	*sep = se;
	return 0;
}


CUTIL_API int ser_element_new_int64 (SERElement **sep, int64_t value)
{
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_INTEGER64 | SER_TYPE_SIGNED;
	se->val.int64 = value;
	se->type |= sizeof (int64_t);
	*sep = se;
	return 0;
}


CUTIL_API int ser_element_new_float (SERElement **sep, float value)
{
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_FLOAT;
	se->val.sfloat = value;
	se->type |= sizeof (float);
	*sep = se;
	return 0;
}


CUTIL_API int ser_element_new_string (SERElement **sep, char *value)
{
	int		length;
	int		mod;
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_STRING;
	length = strlen (value);
	se->val.octets = malloc (length);
	if (se->val.octets == NULL) {
		free (se);
		return ENOMEM;
	}
	strncpy (se->val.octets, value, length);
	se->type |= length;
	mod = length % 4;
	if (mod > 0)
		se->type += ELEMENT_SHIFT_PAD(4-mod);
	*sep = se;
	return 0;
}


CUTIL_API int ser_element_new_string_length (SERElement **sep, char *value, int length)
{
	int		mod;
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_STRING;
	se->val.octets = malloc (length + 1);
	if (se->val.octets == NULL) {
		free (se);
		return ENOMEM;
	}
	strncpy (se->val.octets, value, length);
	se->val.octets[length] = 0;
	se->type |= length;
	mod = length % 4;
	if (mod > 0)
		se->type += ELEMENT_SHIFT_PAD(4-mod);
	*sep = se;
	return 0;
}


CUTIL_API int ser_element_new_binary (SERElement **sep, char *value, int len)
{
	int		mod;
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_BINARY;
	se->val.octets = malloc (len);
	if (se->val.octets == NULL) {
		free (se);
		return ENOMEM;
	}
	memcpy (se->val.octets, value, len);
	se->type |= len;
	mod = len % 4;
	if (mod > 0)
		se->type += ELEMENT_SHIFT_PAD(4-mod);
	*sep = se;
	return 0;
}

/*
int ser_element_new_array (SERElement **sep)
{
	SERElement	*se = NULL;
	if (sep == NULL)
		return EINVAL;
	*sep = NULL;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_ARRAY;
	rc = ser_packet_new (&(SERPacket *)se->val.svoid);
	if (rc != 0) {
		free (se);
		return rc;
	}
	*sep = se;
	return 0;
}
*/

CUTIL_API int ser_element_get_type (SERElement *se, uint32_t *typep)
{
	uint32_t	val;
	if (se == NULL || typep == NULL)
		return EINVAL;
	val = ELEMENT_TYPE (se->type);
	*typep = val;
	return 0;
}


CUTIL_API int ser_element_get_value (SERElement *se, void *valuep, int *lengthp)
{
	int		len;
	char		*p = NULL;
	if (se == NULL || valuep == NULL || lengthp == NULL)
		return EINVAL;
	switch (ELEMENT_TYPE (se->type)) {
		case SER_TYPE_INTEGER16:
			if (*lengthp < 2)
				return ENOSPC;
			*(uint16_t *)valuep = ELEMENT_LENGTH (se->type);
			break;
		case SER_TYPE_INTEGER32:
			if (*lengthp < 4)
				return ENOSPC;
			*(uint32_t *)valuep = se->val.uint32;
			break;
		case SER_TYPE_INTEGER64:
			if (*lengthp < 8)
				return ENOSPC;
			*(uint64_t *)valuep = se->val.uint64;
			break;
		case SER_TYPE_FLOAT:
			if (*lengthp < 4)
				return ENOSPC;
			*(float *)valuep = se->val.sfloat;
			break;
		case SER_TYPE_STRING:
			p = valuep;
			len = *lengthp;
			if (len < (ELEMENT_LENGTH (se->type)+1)) {
				memcpy (p, se->val.octets, len-1);
				p[len] = 0;
			} else {
				len = ELEMENT_LENGTH (se->type);
				memcpy (p, se->val.octets, len);
				/* p[len+1] = 0; */
				p[len] = 0;
			}
			*lengthp = len;
			break;
		case SER_TYPE_BINARY:
			p = valuep;
			len = *lengthp;
			if (len < ELEMENT_LENGTH (se->type)) {
				memcpy (p, se->val.octets, len);
			} else {
				len = ELEMENT_LENGTH (se->type);
				memcpy (p, se->val.octets, len);
			}
			*lengthp = len;
			break;
		case SER_TYPE_ARRAY:
			if (*lengthp < sizeof (void *))
				return ENOSPC;
			*(SERPacket **)valuep = se->val.svoid;
			break;
		default:
			return EINVAL;
	}
	return 0;
}


CUTIL_API int ser_packet_new (SERPacket **spp)
{
	SERPacket	*sp = NULL;
	if (spp == NULL)
		return EINVAL;
	*spp = NULL;
	sp = calloc (1, sizeof (SERPacket));
	if (sp == NULL)
		return ENOMEM;
	*spp = sp;
	return 0;
}


CUTIL_API int ser_packet_free (SERPacket **spp)
{
	int		rc;
	SERPacket	*sp = NULL;
	if (spp == NULL || *spp == NULL)
		return EINVAL;
	sp = *spp;
	rc = ser_packet_clear (sp);
	free (sp);
	*spp = NULL;
	return 0;
}


CUTIL_API int ser_packet_clear (SERPacket *sp)
{
	SERElement	*se = NULL;
	SERElement	*sen = NULL;
	if (sp == NULL)
		return EINVAL;
	se = sp->head;
	while (se != NULL) {
		sen = se->next;
		ser_element_free (&se);
		se = sen;
	}
	se = sp->cache;
	while (se != NULL) {
		sen = se->next;
		ser_element_free (&se);
		se = sen;
	}
	sp->cache = NULL;
	sp->head = NULL;
	sp->tail = NULL;
	sp->iter = NULL;
	sp->elements = 0;
	sp->length = 0;
	sp->flags = 0;
	return 0;
}


CUTIL_API int ser_packet_append (SERPacket *sp, SERElement *se)
{
	if (sp == NULL || se == NULL)
		return EINVAL;
	if (sp->flags & SER_PACKET_FLAG_ARRAYLOCKED)
		return EPERM;
	if (sp->head == NULL) {
		sp->head = se;
	} else {
		if (sp->tail != NULL) {
			sp->tail->next = se;
			se->prev = sp->tail;
		}
	}
	sp->tail = se;
	sp->elements++;
	if (se->type & SER_TYPE_INTEGER16) 
		sp->length += sizeof (uint32_t);
	else
		sp->length += sizeof (uint32_t) + ELEMENT_LENGTH(se->type) + ELEMENT_PAD(se->type);
	return 0;
}


CUTIL_API int ser_packet_prepend (SERPacket *sp, SERElement *se)
{
	if (sp == NULL || se == NULL)
		return EINVAL;
	if (sp->flags & SER_PACKET_FLAG_ARRAYLOCKED)
		return EPERM;
	if (sp->head == NULL) {
		sp->tail = se;
	} else {
		se->next = sp->head;
		sp->head->prev = se;
	}	
	sp->head = se;
	sp->elements++;
	if (se->type & SER_TYPE_INTEGER16) 
		sp->length += sizeof (uint32_t);
	else
		sp->length += sizeof (uint32_t) + ELEMENT_LENGTH(se->type) + ELEMENT_PAD(se->type);
	return 0;
}


CUTIL_API int ser_packet_serialize (SERPacket *sp, char *buffer, int *lenp)
{
	int		rc;
	int		len;
	uint32_t	len32;
	char		*p = NULL;
	SERElement	*se = NULL;
	if (sp == NULL || buffer == NULL || lenp == NULL)
		return EINVAL;
	if (*lenp < sp->length)
		return ENOSPC;
	p = buffer;
	se = sp->head;
	while (se != NULL) {
		*(uint32_t *)p = htonl (se->type);
		p += sizeof (uint32_t);
		if (se->type & SER_TYPE_INTEGER32) {
			*(uint32_t *)p = htonl (se->val.uint32);
			p += sizeof (uint32_t);
		} else if (se->type & SER_TYPE_INTEGER64) {
			*(uint64_t *)p = hton64 (se->val.uint64);
			p += sizeof (uint64_t);
		} else if (se->type & SER_TYPE_FLOAT) {
			*(uint32_t *)p = htonl ((uint32_t)se->val.sfloat);
			p += sizeof (float);
		} else if (se->type & (SER_TYPE_STRING|SER_TYPE_BINARY)) {
			len = ELEMENT_LENGTH(se->type);
			memcpy (p, se->val.octets, len);
			p += len;
			memset (p, 0, ELEMENT_PAD(se->type));
			p += ELEMENT_PAD(se->type);
		} else if (se->type & SER_TYPE_ARRAY) {
			rc = ser_packet_length ((SERPacket *)se->val.svoid, 
						&len32);
			*(uint32_t *)p = htonl (len32);
			p += sizeof (uint32_t);
			rc = ser_packet_serialize ((SERPacket *)se->val.svoid,
						p,
						&len32);
			if (rc != 0) {
				*lenp = 0;
				return rc;
			}
			p += len32;
		} else if (se->type & SER_TYPE_INTEGER16) {
			/* do nothing else */
		} else {
			/* Invalid element type */
			*lenp = 0;
			return EINVAL;
		}
		se = se->next;
	}
	len = p - buffer;
	*lenp = len;
	return 0;
}


CUTIL_API int ser_packet_deserialize (SERPacket **spp, char *buffer, int len)
{
	int		l;
	int		rc;
	uint32_t	type;
	uint16_t	s;
	char		*p = NULL;
	char		*bend = NULL;
	SERElement	*se = NULL;
	SERPacket	*sp = NULL;
	SERPacket	*spa = NULL;
	if (spp == NULL || buffer == NULL)
		return EINVAL;
	*spp = NULL;
	if (len < 4)
		return EINVAL;
	rc = ser_packet_new (&sp);
	if (rc != 0)
		return rc;
	p = buffer;
	bend = buffer + len;
	while (p < bend) {
		type = *(uint32_t *)p;
		type = ntohl (type);
		p += sizeof (uint32_t);
		if (ELEMENT_TYPE(type) & SER_TYPE_INTEGER16) {
			s = ELEMENT_LENGTH (type);
			rc = ser_element_new (&se,
						ELEMENT_TYPE(type),
						(void *)&s,
						sizeof (uint16_t));
			if (rc != 0) {
				ser_packet_free (&sp);
				return rc;
			}
			rc = ser_packet_append (sp, se);
			if (rc != 0) {
				ser_packet_free (&sp);
				return rc;
			}
		} else if (ELEMENT_TYPE(type) & SER_TYPE_ARRAY) {
				*(uint32_t*)p = ntohl (*(uint32_t*)p);
				l = *(uint32_t*)p;
				p += sizeof (uint32_t);
				rc = ser_packet_deserialize (&spa,
							p,
							l);
				if (rc != 0) {
					ser_packet_free (&sp);
					return rc;
				}
				rc = ser_packet_append_as_array (sp, spa);
				if (rc != 0) {
					ser_packet_free (&spa);
					ser_packet_free (&sp);
					return rc;
				}
				p += l;
		} else {
			if (ELEMENT_TYPE (type) & SER_TYPE_INTEGER32)
				*(uint32_t*)p = ntohl (*(uint32_t*)p);
			else if (ELEMENT_TYPE (type) & SER_TYPE_INTEGER64)
				*(uint64_t*)p = ntoh64 (*(uint64_t*)p);
			else if (ELEMENT_TYPE (type) & SER_TYPE_FLOAT)
				*(uint32_t*)p = ntohl (*(uint32_t*)p);
			l = ELEMENT_LENGTH (type);
			rc = ser_element_new (&se,
						ELEMENT_TYPE(type),
						(void *)p,
						l);
			if (rc != 0) {
				ser_packet_free (&sp);
				return rc;
			}
			rc = ser_packet_append (sp, se);
			if (rc != 0) {
				ser_packet_free (&sp);
				return rc;
			}
			p += l;
			p += ELEMENT_PAD (type);
		}
	}
	*spp = sp;
	return 0;
}


CUTIL_API SERElement *ser_packet_iterate_elements (SERPacket *sp, uint32_t whence)
{
	SERElement	*se = NULL;
	if (sp == NULL)
		return NULL;
	switch (whence) {
		case SER_WHENCE_FIRST:
			sp->iter = sp->head;
			break;
		case SER_WHENCE_LAST:
			sp->iter = sp->tail;
			break;
		case SER_WHENCE_NEXT:
			sp->iter = sp->iter->next;
			break;
		case SER_WHENCE_PREVIOUS:
			sp->iter = sp->iter->prev;
			break;
		default:
			return NULL;
	}
	se = sp->iter;
	return se;
}


CUTIL_API int ser_packet_length (SERPacket *sp, uint32_t *lengthp)
{
	if (sp == NULL || lengthp == NULL)
		return EINVAL;
	*lengthp = sp->length;
	return 0;
}


CUTIL_API int ser_packet_elements (SERPacket *sp, uint32_t *countp)
{
	if (sp == NULL || countp == NULL)
		return EINVAL;
	*countp = sp->elements;
	return 0;
}


CUTIL_API int ser_packet_append_as_array (SERPacket *sp, SERPacket *ssp)
{
	int		rc;
	uint32_t	len;
	SERElement	*se = NULL;
	if (sp == NULL || ssp == NULL)
		return EINVAL;
	if (sp->flags & SER_PACKET_FLAG_ARRAYLOCKED)
		return EPERM;
	se = calloc (1, sizeof (SERElement));
	if (se == NULL)
		return ENOMEM;
	se->type = SER_TYPE_ARRAY;
	se->type |= sizeof (uint32_t);
	se->val.svoid = ssp;
	if (sp->head == NULL) {
		sp->head = se;
	} else {
		if (sp->tail != NULL) {
			sp->tail->next = se;
			se->prev = sp->tail;
		}
	}
	sp->tail = se;
	sp->elements++;
	rc = ser_packet_length (ssp, &len);
	sp->length += sizeof (uint32_t) + ELEMENT_LENGTH(se->type) + ELEMENT_PAD(se->type) + len;
	ssp->flags |= SER_PACKET_FLAG_ARRAYLOCKED;
	return 0;
}


CUTIL_API int ser_packet_new_with_format (SERPacket **spp, char *format, ...)
{
	va_list		varg;
	int		rc;
	int16_t		i16;
	uint16_t	u16;
	int32_t		i32;
	uint32_t	u32;
	int64_t		i64;
	uint64_t	u64;
	char		*s;
	int		len;	
	SERPacket	*sp;
	SERElement	*se;

	va_start(varg, format);
	if (spp == NULL){
		va_end(varg);
		return EINVAL;
	}
	*spp = NULL;
	if (format == NULL){
		va_end(varg);
		return EINVAL;
	}
	rc = ser_packet_new(&sp);
	if (rc)
		return rc;
	while (*format) {
		switch(*format++) {
			case 's':
				i16 = va_arg(varg, int);
				rc = ser_element_new_int16(&se, i16);
				break;
			case 'S':
				u16 = va_arg(varg, unsigned int);
				rc = ser_element_new_uint16(&se, u16);
				break;
			case 'i':
				i32 = va_arg(varg, int32_t);
				rc = ser_element_new_int32(&se, i32);
				break;
			case 'I':
				u32 = va_arg(varg, uint32_t);
				rc = ser_element_new_uint32(&se, u32);
				break;
			case 'l':
				i64 = va_arg(varg, int64_t);
				rc = ser_element_new_int64(&se, i64);
				break;
			case 'L':
				u64 = va_arg(varg, uint64_t);
				rc = ser_element_new_uint64(&se, u64);
				break;
			case 'c':
				s = va_arg(varg, char*);
				rc = ser_element_new_string(&se, s);
				break;
			case 'C':
				s = va_arg(varg, char*);
				len = va_arg(varg, int);
				rc = ser_element_new_string_length(&se, s, len);
				break;
			case 'b':
				s = va_arg(varg, char*);
				len = va_arg(varg, int);
				rc = ser_element_new_binary(&se, s, len);
				break;
			default:
				rc = EINVAL;
				break;
		}
		if (rc) {
			ser_packet_free(&sp);
			break;
		}
		rc = ser_packet_append(sp, se);
		if (rc) {
			ser_packet_free(&sp);
			break;
		}
	}

	va_end(varg);
	*spp = sp;
	return rc;
}



CUTIL_API int ser_packet_get_elements_with_format (SERPacket *sp, char *format, ...)
{
	va_list		varg;
	int		rc;
	int16_t		*pi16;
	uint16_t	*pu16;
	int32_t		*pi32;
	uint32_t	*pu32;
	int64_t		*pi64;
	uint64_t	*pu64;
	char		*s;
	int		len;
	int		*plen;
	SERElement	*se;

	va_start(varg, format);
	if (sp == NULL){
		va_end(varg);
		return EINVAL;
	}

	rc = 0;
	se = ser_packet_iterate_elements(sp, SER_WHENCE_FIRST);
	while (*format && se) {
		switch(*format++) {
			case 's':
				pi16 = va_arg(varg, int*);
				len = 2;
				rc = ser_element_get_value(se, pi16, &len);
				break;
			case 'S':
				pu16 = va_arg(varg, unsigned int*);
				len = 2;
				rc = ser_element_get_value(se, pu16, &len);
				break;
			case 'i':
				pi32 = va_arg(varg, int32_t*);
				len = 4;
				rc = ser_element_get_value(se, pi32, &len);
				break;
			case 'I':
				pu32 = va_arg(varg, uint32_t*);
				len = 4;
				rc = ser_element_get_value(se, pu32, &len);
				break;
			case 'l':
				pi64 = va_arg(varg, int64_t*);
				len = 8;
				rc = ser_element_get_value(se, pi64, &len);
				break;
			case 'L':
				pu64 = va_arg(varg, uint64_t*);
				len = 8;
				rc = ser_element_get_value(se, pu64, &len);
				break;
/*
			case 'c':
				s = va_arg(varg, char*);
				*plen = 0;
				rc = ser_element_get_value(&se, s, plen);
				break;
*/
			case 'C':
				s = va_arg(varg, char*);
				plen = va_arg(varg, int*);
				rc = ser_element_get_value(se, s, plen);
				break;
			case 'b':
				s = va_arg(varg, char*);
				plen = va_arg(varg, int*);
				rc = ser_element_get_value(se, s, plen);
				break;
			default:
				rc = EINVAL;
				break;
		}
		if (rc)
			break;
		se = ser_packet_iterate_elements(sp, SER_WHENCE_NEXT);
	}
	va_end(varg);
	return rc;
}



/* Caching functions */
CUTIL_API int ser_packet_new_element_c (SERPacket *sp, SERElement **sep,
                                int type, void *valuep, int length)
{
	int		mod;
	SERElement	*se = NULL;
	if (sp == NULL || sep == NULL || valuep == NULL)
		return EINVAL;
	if (sp->flags & SER_PACKET_FLAG_ARRAYLOCKED)
		return EPERM;
	*sep = NULL;
	if (sp->cache != NULL) {
		se = sp->cache;
		sp->cache = se->next;
		/*sp->cache->prev = NULL; */
		se->refs = 0;
		se->next = NULL;
		se->prev = NULL;
		se->type = 0;
	} else {
		se = calloc (1, sizeof (SERElement));
		if (se == NULL)
			return ENOMEM;
	}
		
	if (type & SER_TYPE_INTEGER16) {
		se->type = SER_TYPE_INTEGER16;
		if (type & SER_TYPE_SIGNED)
			se->type |= SER_TYPE_SIGNED;
		se->type |= *(uint16_t *)valuep;
	} else if (type & SER_TYPE_INTEGER32) {
		se->type = SER_TYPE_INTEGER32;
		if (type & SER_TYPE_SIGNED)
			se->type |= SER_TYPE_SIGNED;
		se->val.uint32 = *(uint32_t *)valuep;
		se->type |= sizeof (uint32_t);
	} else if (type & SER_TYPE_INTEGER64) {
		se->type = SER_TYPE_INTEGER64;
		if (type & SER_TYPE_SIGNED)
			se->type |= SER_TYPE_SIGNED;
		se->val.uint64 = *(uint64_t *)valuep;
		se->type |= sizeof (uint64_t);
	} else if (type & SER_TYPE_FLOAT) {
		se->type = SER_TYPE_FLOAT;
		se->val.sfloat = *(float *)valuep;
		se->type |= sizeof (float);
	} else if (type & SER_TYPE_STRING) {
		se->type = SER_TYPE_STRING;
		se->val.octets = malloc (length + 1);
		if (se->val.octets == NULL) {
			se->next = sp->cache;
			sp->cache = se;
			return ENOMEM;
		}
		strncpy (se->val.octets, valuep, length);
		se->val.octets[length] = 0;
		se->type |= length;
		mod = length % 4;
		if (mod > 0)
			se->type += ELEMENT_SHIFT_PAD(4-mod);
	} else if (type & SER_TYPE_BINARY) {
		se->type = SER_TYPE_BINARY;
		se->val.octets = malloc (length);
		if (se->val.octets == NULL) {
			se->next = sp->cache;
			sp->cache = se;
			return ENOMEM;
		}
		memcpy (se->val.octets, valuep, length);
		se->type |= length;
		mod = length % 4;
		if (mod > 0)
			se->type += ELEMENT_SHIFT_PAD(4-mod);
	} else {
		se->next = sp->cache;
		sp->cache = se;
		return EINVAL;
	}
	*sep = se;
	return 0;
}


CUTIL_API int ser_packet_deserialize_c (SERPacket *sp, char *buffer, int len)
{
	int		l;
	int		rc;
	uint32_t	type;
	uint16_t	s;
	char		*p = NULL;
	char		*bend = NULL;
	SERElement	*se = NULL;
	SERPacket	*spa = NULL;
	if (sp == NULL || buffer == NULL)
		return EINVAL;
	if (len < 4)
		return EINVAL;
	p = buffer;
	bend = buffer + len;
	while (p < bend) {
		type = *(uint32_t *)p;
		type = ntohl (type);
		p += sizeof (uint32_t);
		if (ELEMENT_TYPE(type) & SER_TYPE_INTEGER16) {
			s = ELEMENT_LENGTH (type);
			rc = ser_packet_new_element_c (sp,
						&se,
						ELEMENT_TYPE(type),
						(void *)&s,
						sizeof (uint16_t));
			if (rc != 0) {
				ser_packet_clear_c (sp);
				return rc;
			}
			rc = ser_packet_append (sp, se);
			if (rc != 0) {
				ser_element_free (&se);
				ser_packet_clear_c (sp);
				return rc;
			}
		} else if (ELEMENT_TYPE(type) & SER_TYPE_ARRAY) {
				*(uint32_t*)p = ntohl (*(uint32_t*)p);
				l = *(uint32_t*)p;
				p += sizeof (uint32_t);
				rc = ser_packet_deserialize (&spa,
							p,
							l);
				if (rc != 0) {
					ser_packet_clear_c (sp);
					return rc;
				}
				rc = ser_packet_append_as_array (sp, spa);
				if (rc != 0) {
					ser_packet_free (&spa);
					ser_packet_clear_c (sp);
					return rc;
				}
				p += l;
		} else {
			if (ELEMENT_TYPE (type) & SER_TYPE_INTEGER32)
				*(uint32_t*)p = htonl (*(uint32_t*)p);
			else if (ELEMENT_TYPE (type) & SER_TYPE_INTEGER64)
				*(uint64_t*)p = ntoh64 (*(uint64_t*)p);
			else if (ELEMENT_TYPE (type) & SER_TYPE_FLOAT)
				*(uint32_t*)p = htonl (*(uint32_t*)p);
			l = ELEMENT_LENGTH (type);
			rc = ser_packet_new_element_c (sp,
						&se,
						ELEMENT_TYPE(type),
						(void *)p,
						l);
			if (rc != 0) {
				ser_packet_clear_c (sp);
				return rc;
			}
			rc = ser_packet_append (sp, se);
			if (rc != 0) {
				ser_element_free (&se);
				ser_packet_clear_c (sp);
				return rc;
			}
			p += l;
			p += ELEMENT_PAD (type);
		}
	}
	return 0;
}




CUTIL_API int ser_packet_clear_c (SERPacket *sp)
{
	SERElement	*se = NULL;
	if (sp == NULL)
		return EINVAL;
	if (sp->head == NULL)
		return 0;
	se = sp->head;
	while (se != NULL) {
		if (se->type & (SER_TYPE_STRING|SER_TYPE_BINARY)) {
			free (se->val.octets);
			se->val.octets = NULL;
		} else if (se->type & SER_TYPE_ARRAY) {
			ser_packet_free ((SERPacket **)&se->val.svoid);
			se->val.svoid = NULL;
		}
		se = se->next;
	}
	sp->cache = sp->head;
	sp->head = NULL;
	sp->tail = NULL;
	sp->iter = NULL;
	sp->elements = 0;
	sp->length = 0;
	sp->flags = 0;
	return 0;
}




