/************************************************************************/
/* url.c - url construct/parse functions source				*/
/************************************************************************/
/*
 *  (c)2003 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <cutil/strfuncs.h>
#include <cutil/attribute.h>
#include <cutil/url.h>

#ifdef	DMALLOC
#include <dmalloc.h>
#endif


#define URL_ALLOC_SIZE		256
#define URL_PARAMS_MAX		32

typedef struct _LString	LString;
struct _LString {
	char *buffer;
	int alloclen;
};


CUTIL_API int url_attrib_append (Attribute *attrib, LString *lstring);

CUTIL_API int url_new (Url **urlp, char *protocol, char *hostname,
		char *path, int port)
{
	Url *url = NULL;
	if (urlp == NULL || protocol == NULL || hostname == NULL)
		return EINVAL;
	url = (Url *)calloc (1, sizeof (Url));
	if (url == NULL)
		return ENOMEM;
	*urlp = url;
	url->protocol = strdup (protocol);
	url->hostname = strdup (hostname);
	if (path != NULL)
		url->path = strdup (path);
	url->port = port;
	return 0;
}


CUTIL_API int url_add_param (Url *url, char *name, char *value)
{
	if (url == NULL || name == NULL || value == NULL)
		return EINVAL;
	if (attribute_find (url->attrib, name) != NULL)
		return EEXIST;
	url->attrib = attribute_add (url->attrib, name, value);
	return 0;
}

CUTIL_API int url_construct (Url *url, char **stringp)
{
	int rc;
	int len;
	LString str;
	if (url == NULL || stringp == NULL)
		return EINVAL;
	*stringp = NULL;
	len = URL_ALLOC_SIZE;
	str.buffer = calloc (1, len);
	if (str.buffer == NULL)
		return ENOMEM;
	str.alloclen = len;
	if (url->port == 0) {
		rc = snprintf (str.buffer,
				str.alloclen,
				"%s://%s/",
				url->protocol,
				url->hostname);
	} else {
		rc = snprintf (str.buffer,
				str.alloclen,
				"%s://%s:%d/",
				url->protocol,
				url->hostname,
				url->port);
	}
	if (rc > str.alloclen) {
		free (str.buffer);
		return EOVERFLOW;
	}

	if (url->path == NULL) {
		*stringp = str.buffer;
		return 0; /* no path */
	}
	len = strlen (str.buffer) + strlen (url->path) + 1;
	if (len > str.alloclen) {
		free (str.buffer);
		return EOVERFLOW;
	}
	strcat (str.buffer, url->path);

	if (url->attrib == NULL) {
		*stringp = str.buffer;
		return 0; /* no params */
	}
	len = strlen (str.buffer) + 1 + 1;
	if (len > str.alloclen) {
		free (str.buffer);
		return EOVERFLOW;
	}
	strcat (str.buffer, "?");
	rc = attribute_foreach (url->attrib,
				(AttributeCallbackFunction)url_attrib_append,
				&str);
	if (rc != 0) {
		free (str.buffer);
		return rc;
	}
	*stringp = str.buffer;
	return 0;
}


CUTIL_API int url_new_parse (Url **urlp, char *string)
{
	int rc;
	int i;
	int items;
	int count;
	struct servent *se = NULL;
	char **pitem = NULL;
	char **hitem = NULL;
	Url *url = NULL;
	if (urlp == NULL || string == NULL)
		return EINVAL;
	*urlp = NULL;
	url = (Url *)calloc (1, sizeof (Url));
	if (url == NULL)
		return ENOMEM;

	/* protocol */
	items = 2;
	rc = strnsplit (string, strlen (string), ':', &pitem, &items);
	if (rc != 0) {
		url_free (&url);
		return rc;
	}
	url->protocol = strdup (pitem[0]);

	/* hostname and port */
	items = 2;
	rc = strnsplit (pitem[1], strlen (pitem[1]), '/', &hitem, &items);
	if (rc != 0) {
		strarray_free (&pitem);
		url_free (&url);
		return rc;
	}
	strarray_free (&pitem);
	items = 2;
	rc = strnsplit (hitem[0], strlen (hitem[0]), ':', &pitem, &items);
	if (rc != 0) {
		strarray_free (&hitem);
		url_free (&url);
		return rc;
	}
	url->hostname = strdup (pitem[0]);
	if (pitem[1] != NULL) {
		i = atoi (pitem[1]);
		if (i == 0) {
			se = getservbyname (pitem[1], "tcp");
			if (se != NULL) {
				url->port = ntohs (se->s_port);
			} else {
				url->port = 0;
			}
		} else {
			url->port = i;
		}
	}
	strarray_free (&pitem);
	if (hitem[1] == NULL) {
		strarray_free (&hitem);
		*urlp = url;
		return 0; /* no path to parse */
	}

	/* path */
	items = 2;
	rc = strnsplit (hitem[1], strlen (hitem[1]), '?', &pitem, &items);
	if (rc != 0) {
		strarray_free (&hitem);
		url_free (&url);
		return rc;
	}
	url->path = strdup (pitem[0]);
	strarray_free (&hitem);
	if (pitem[1] == NULL) {
		strarray_free (&pitem);
		*urlp = url;
		return 0; /* no params to parse */
	}

	/* params */
	items = URL_PARAMS_MAX;
	rc = strnsplit (pitem[1], strlen (pitem[1]), '&', &hitem, &items);
	if (rc != 0) {
		strarray_free (&pitem);
		url_free (&url);
		return rc;
	}
	strarray_free (&pitem);
	count = items;
	for (i=0 ; i<count ; i++) {
		items = 2;
		rc = strnsplit (hitem[i], strlen (hitem[i]),
				'=', &pitem, &items);
		if (rc != 0) {
			strarray_free (&hitem);
			url_free (&url);
			return rc;
		}
		url->attrib = attribute_add (url->attrib, pitem[0], pitem[1]);
		strarray_free (&pitem);
	}
	strarray_free (&hitem);
	*urlp = url;
	return 0;
}


CUTIL_API int url_get_protocol (Url *url, char **protocol)
{
	if (url == NULL || protocol == NULL)
		return EINVAL;
	if (url->protocol == NULL)
		*protocol = "";
	else
		*protocol = url->protocol;
	return 0;
}


CUTIL_API int url_get_hostname (Url *url, char **hostname)
{
	if (url == NULL || hostname == NULL)
		return EINVAL;
	if (url->hostname == NULL)
		*hostname = "";
	else
		*hostname = url->hostname;
	return 0;
}


CUTIL_API int url_get_path (Url *url, char **path)
{
	if (url == NULL || path == NULL)
		return EINVAL;
	if (url->path == NULL)
		*path = "";
	else
		*path = url->path;
	return 0;
}


CUTIL_API int url_get_port (Url *url, int *port)
{
	if (url == NULL || port == NULL)
		return EINVAL;
	*port = url->port;
	return 0;
}


CUTIL_API int url_has_param (Url *url, char *name)
{
	if (url == NULL || name == NULL)
		return EINVAL;
	if (attribute_find (url->attrib, name) != NULL)
		return 1;
	return 0;
}


CUTIL_API int url_get_param (Url *url, char *name, char **value)
{
	if (url == NULL || name == NULL || value == NULL)
		return EINVAL;
	*value = attribute_get_value (url->attrib, name);
	if (*value == NULL)
		return ENOENT;
	return 0;
}


CUTIL_API int url_free (Url **urlp)
{
	Url *url = NULL;
	if (urlp == NULL || *urlp == NULL)
		return EINVAL;
	url = *urlp;
	if (url->hostname)
		free (url->hostname);
	if (url->protocol)
		free (url->protocol);
	if (url->path)
		free (url->path);
	attribute_freeall (&url->attrib);
	free (url);
	*urlp = NULL;
	return 0;
}


CUTIL_API int url_attrib_append (Attribute *attrib, LString *lstring)
{
	int rc;
	int len;
	char *name = NULL;
	char *value = NULL;
	if (attrib == NULL || lstring == NULL || lstring->buffer == NULL)
		return EINVAL;
	rc = attribute_get_name_and_value (attrib, &name, &value);
	if (rc != 0)
		return rc;
	len = strlen (lstring->buffer) + strlen (name) + 1 +
		strlen (value) + 1;
	if (len > lstring->alloclen)
		return EOVERFLOW;
	len = strlen (lstring->buffer);
	if (lstring->buffer[len-1] != '?')
		strcat (lstring->buffer, "&");
	strcat (lstring->buffer, name);
	strcat (lstring->buffer, "=");
	strcat (lstring->buffer, value);
	return 0;
}


