/************************************************************************/
/* xmlrpc.c - XMLRPC Implementation Source				*/
/************************************************************************/
/*
 *  (c)2003,2010 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <cutil/xmlrpc_p.h>

#ifdef	DMALLOC
#include <dmalloc.h>
#endif


/* Forward prototypes for these functions, prevents compiler warnings */
CUTIL_API int xmlrpc_struct_ref_dec (XmlrpcStruct **xsp);
CUTIL_API int xmlrpc_array_ref_dec (XmlrpcArray **xap);
CUTIL_API int xmlrpc_struct_new (XmlrpcStruct **xsp);
CUTIL_API int xmlrpc_struct_add_member (XmlrpcStruct *xs, char *membername,
				XmlrpcValue *membervalue);
CUTIL_API int xmlrpc_request_free (XmlrpcRequest **xrp);

/* Forward prototypes for internal functions */
int xmlrpc_value_serialize_xml (xmlNodePtr *nodep, XmlrpcValue *xv);
int xmlrpc_value_new_parse_xml (XmlrpcValue **xvp, xmlNodePtr valuenode);
int xmlrpc_struct_new_parse_xml (XmlrpcStruct **xsp, xmlNodePtr structnode);
int xmlrpc_array_new_parse_xml (XmlrpcArray **xap, xmlNodePtr arraynode);


/************************************************************************/
/* XMLRPC Reference Counting functions */
/************************************************************************/
CUTIL_API int xmlrpc_value_ref_inc (XmlrpcValue *xv)
{
	if (xv == NULL)
		return EINVAL;
	xv->ref.refcount++;
	return 0;
}


CUTIL_API int xmlrpc_value_ref_dec (XmlrpcValue **xvp)
{
	XmlrpcValue *xv = NULL;
	if (xvp == NULL || *xvp == NULL)
		return EINVAL;
	xv = *xvp;
	xv->ref.refcount--;
	if (xv->ref.refcount > 0)
		return 0;
	if (xv->data) {
		if (xv->type == XMLRPC_VALUETYPE_STRUCT) {
			xmlrpc_struct_ref_dec ((XmlrpcStruct **)&xv->data);
		} else if (xv->type == XMLRPC_VALUETYPE_ARRAY) {
			xmlrpc_array_ref_dec ((XmlrpcArray **)&xv->data);
		} else{
			free (xv->data);
		}
	}
	free (xv);
	*xvp = NULL;
	return 0;
}


CUTIL_API int xmlrpc_struct_ref_inc (XmlrpcStruct *xs)
{
	if (xs == NULL)
		return EINVAL;
	xs->ref.refcount++;
	return 0;
}


CUTIL_API int xmlrpc_struct_ref_dec (XmlrpcStruct **xsp)
{
	SList *litem;
	XmlrpcStructMember *xsm = NULL;
	XmlrpcStruct *xs = NULL;
	if (xsp == NULL || *xsp == NULL)
		return EINVAL;
	xs = *xsp;

	xs->ref.refcount--;
	if (xs->ref.refcount > 0)
		return 0;

	if (xs->memberlist != NULL) {
		litem = xs->memberlist;
		while (litem != NULL) {
			xsm = slist_get_data (litem);
			xmlrpc_value_ref_dec (&xsm->value);
			free (xsm->name);
			free (xsm);
			litem = slist_get_next (litem);
		}
		slist_free (&xs->memberlist);
	}
	free (xs);
	*xsp = NULL;
	return 0;
}


CUTIL_API int xmlrpc_array_ref_inc (XmlrpcArray *xa)
{
	if (xa == NULL)
		return EINVAL;
	xa->ref.refcount++;
	return 0;
}


CUTIL_API int xmlrpc_array_ref_dec (XmlrpcArray **xap)
{
	int rc;
	SList *litem;
	XmlrpcValue *xv = NULL;
	XmlrpcArray *xa = NULL;
	if (xap == NULL || *xap == NULL)
		return EINVAL;
	xa = *xap;

	xa->ref.refcount--;
	if (xa->ref.refcount > 0)
		return 0;

	if (xa->valuelist != NULL) {
		litem = xa->valuelist;
		while (litem != NULL) {
			xv = slist_get_data (litem);
			rc = xmlrpc_value_ref_dec (&xv);
			litem = slist_get_next (litem);
		}
		slist_free (&xa->valuelist);
	}
	free (xa);
	*xap = NULL;
	return 0;
}


/************************************************************************/
/* XMLRPC value functions */
/************************************************************************/
CUTIL_API int xmlrpc_value_new (XmlrpcValue **xvp, int valuetype,
			void *value, int valuelen)
{
	int rc;
	char *p;
	XmlrpcValue *xv = NULL;

	if (xvp == NULL)
		return EINVAL;
	*xvp = NULL;
	if (value == NULL || valuelen < 0)
		return EINVAL;
	if (valuetype < XMLRPC_VALUETYPE_MIN || valuetype > XMLRPC_VALUETYPE_MAX)
		return EINVAL;
	xv = (XmlrpcValue *)calloc(1, sizeof (XmlrpcValue));
	if (xv == NULL)
		return ENOMEM;
	rc = xmlrpc_value_ref_inc (xv);
	xv->type = valuetype;
	xv->datalen = valuelen;
	if (valuetype == XMLRPC_VALUETYPE_STRUCT) {
		xv->data = value;
		xmlrpc_struct_ref_inc ((XmlrpcStruct *)value);
	} else if (valuetype == XMLRPC_VALUETYPE_ARRAY) {
		xv->data = value;
		xmlrpc_array_ref_inc ((XmlrpcArray *)value);
	} else {
		xv->data = calloc (1, valuelen+1);
		if (xv->data == NULL) {
			free (xv);
			return ENOMEM;
		}
		memcpy (xv->data, value, valuelen);
		p = (char *)xv->data;
		p[valuelen] = 0;
	}
	*xvp = xv;
	return 0;
}


CUTIL_API int xmlrpc_value_new_from_struct (XmlrpcValue **xvp, XmlrpcStruct *xs)
{
	int rc;
	XmlrpcValue *xv = NULL;

	if (xvp == NULL)
		return EINVAL;
	*xvp = NULL;
	if (xs == NULL)
		return EINVAL;
	rc = xmlrpc_value_new (&xv,
			XMLRPC_VALUETYPE_STRUCT,
			xs,
			sizeof (XmlrpcStruct));
	if (rc != 0)
		return rc;
	*xvp = xv;
	return 0;
}


CUTIL_API int xmlrpc_value_new_from_array (XmlrpcValue **xvp, XmlrpcArray *xa)
{
	int rc;
	XmlrpcValue *xv = NULL;

	if (xvp == NULL)
		return EINVAL;
	*xvp = NULL;
	if (xa == NULL)
		return EINVAL;
	rc = xmlrpc_value_new (&xv,
			XMLRPC_VALUETYPE_ARRAY,
			xa,
			sizeof (XmlrpcArray));
	if (rc != 0)
		return rc;
	*xvp = xv;
	return rc;
}

CUTIL_API int xmlrpc_value_new_fault (XmlrpcValue **xvp,
			int faultcode, char *faultstring)
{
	int rc;
	XmlrpcStruct *xs = NULL;
	XmlrpcValue *xv = NULL;
	if (xvp == NULL || faultstring == NULL)
		return EINVAL;
	*xvp = NULL;
	rc = xmlrpc_struct_new (&xs);
	if (rc != 0)
		return rc;

	rc = xmlrpc_value_new (&xv,
				XMLRPC_VALUETYPE_INT,
				&faultcode, 
				sizeof (faultcode));
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		return rc;
	}
	rc = xmlrpc_struct_add_member (xs, "faultCode", xv);
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		xmlrpc_value_ref_dec (&xv);
		return rc;
	}
	rc = xmlrpc_value_ref_dec (&xv);

	rc = xmlrpc_value_new (&xv,
				XMLRPC_VALUETYPE_STRING,
				faultstring, 
				strlen (faultstring));
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		return rc;
	}
	rc = xmlrpc_struct_add_member (xs, "faultString", xv);
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		xmlrpc_value_ref_dec (&xv);
		return rc;
	}
	rc = xmlrpc_value_ref_dec (&xv);

	rc = xmlrpc_value_new_from_struct (&xv, xs);
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		return rc;
	}
	rc = xmlrpc_struct_ref_dec (&xs);
	*xvp = xv;
	return 0;
}



CUTIL_API int xmlrpc_value_get_value (XmlrpcValue *xv, int *valuetype, void **valuep,
			int *valuelen)
{
	if (xv == NULL || valuetype == NULL)
		return EINVAL;
	if (valuep == NULL || valuelen == NULL)
		return EINVAL;
	*valuetype = 0;
	*valuep = NULL;
	*valuelen = 0;
	if (xv->type < XMLRPC_VALUETYPE_MIN || xv->type > XMLRPC_VALUETYPE_MAX)
		return EINVAL;
	*valuetype = xv->type;
	*valuep = xv->data;
	*valuelen = xv->datalen;
	return 0;
}


CUTIL_API int xmlrpc_value_get_struct (XmlrpcValue *xv, XmlrpcStruct **xsp)
{
	if (xv == NULL || xsp == NULL)
		return EINVAL;
	*xsp = NULL;
	if (xv->type != XMLRPC_VALUETYPE_STRUCT)
		return EINVAL;
	*xsp = (XmlrpcStruct *)xv->data;
	return 0;
}


CUTIL_API int xmlrpc_value_get_array (XmlrpcValue *xv, XmlrpcArray **xap)
{
	if (xv == NULL || xap == NULL)
		return EINVAL;
	*xap = NULL;
	if (xv->type != XMLRPC_VALUETYPE_ARRAY)
		return EINVAL;
	*xap = (XmlrpcArray *)xv->data;
	return 0;
}



/************************************************************************/
/* XMLRPC struct functions */
/************************************************************************/
CUTIL_API int xmlrpc_struct_new (XmlrpcStruct **xsp)
{
	int rc;
	XmlrpcStruct *xs = NULL;
	if (xsp == NULL)
		return EINVAL;
	*xsp = NULL;
	xs = (XmlrpcStruct *)calloc (1, sizeof (XmlrpcStruct));
	if (xs == NULL)
		return ENOMEM;
	rc = xmlrpc_struct_ref_inc (xs);
	*xsp = xs;
	return 0;
}


CUTIL_API int xmlrpc_struct_add_member (XmlrpcStruct *xs, char *membername,
				XmlrpcValue *membervalue)
{
	int rc;
	XmlrpcStructMember *xsm = NULL;
	if (xs == NULL || membername == NULL || membervalue == NULL)
		return EINVAL;
	xsm = (XmlrpcStructMember *)calloc (1, sizeof (XmlrpcStructMember));
	if (xsm == NULL)
		return ENOMEM;
	xsm->name = strdup (membername);
	if (xsm->name == NULL) {
		free (xsm);
		return ENOMEM;
	}
	xsm->value = membervalue;
	rc = slist_item_add (&xs->memberlist, xsm);
	if (rc != 0) {
		free (xsm->name);
		free (xsm);
		return rc;
	}
	rc = xmlrpc_value_ref_inc (membervalue);
	return 0;
}


CUTIL_API int xmlrpc_struct_member_count (XmlrpcStruct *xs, int *membercount)
{
	int rc;
	if (xs == NULL || membercount == NULL)
		return EINVAL;
	*membercount = 0;
	if (xs->memberlist == NULL)
		return 0;
	rc = slist_item_count (xs->memberlist, membercount);
	return rc;
}


CUTIL_API int xmlrpc_struct_get_member (XmlrpcStruct *xs, char *membername,
				XmlrpcValue **membervalue)
{
	SList *litem;
	XmlrpcStructMember *xsm = NULL;
	if (xs == NULL || membername == NULL || membervalue == NULL)
		return EINVAL;
	*membervalue = NULL;

	litem = xs->memberlist;
	while (litem != NULL) {
		xsm = slist_get_data (litem);
		if (!strcmp(xsm->name, membername)) {
			xmlrpc_value_ref_inc (xsm->value);
			*membervalue = xsm->value;
			return 0;
		}
		litem = slist_get_next (litem);
	}
	return ENOENT;
}


CUTIL_API int xmlrpc_struct_has_member (XmlrpcStruct *xs, char *membername, int *hasmember)
{
	SList *litem;
	XmlrpcStructMember *xsm = NULL;
	if (xs == NULL || membername == NULL || hasmember == NULL)
		return EINVAL;
	*hasmember = 0;

	litem = xs->memberlist;
	while (litem != NULL) {
		xsm = slist_get_data (litem);
		if (!strcmp(xsm->name, membername)) {
			*hasmember = 1;
			return 0;
		}
		litem = slist_get_next (litem);
	}
	return 0;
}


CUTIL_API int xmlrpc_struct_remove_member (XmlrpcStruct *xs, char *membername)
{
	int rc;
	SList *litem;
	XmlrpcStructMember *xsm = NULL;
	if (xs == NULL || membername == NULL)
		return EINVAL;

	litem = xs->memberlist;
	while (litem != NULL) {
		xsm = slist_get_data (litem);
		if (!strcmp(xsm->name, membername)) {
			break;
		}
		litem = slist_get_next (litem);
	}
	if (litem == NULL)
		return ENOENT;
	rc = slist_item_remove (&xs->memberlist, xsm);
	rc = xmlrpc_value_ref_dec (&xsm->value);
	free (xsm->name);
	free (xsm);
	return rc;
}



/************************************************************************/
/* XMLRPC array functions */
/************************************************************************/
CUTIL_API int xmlrpc_array_new (XmlrpcArray **xap)
{
	int rc;
	XmlrpcArray *xa = NULL;
	if (xap == NULL)
		return EINVAL;
	*xap = NULL;
	xa = (XmlrpcArray *)calloc (1, sizeof (XmlrpcArray));
	if (xa == NULL)
		return ENOMEM;
	rc = xmlrpc_array_ref_inc (xa);
	*xap = xa;
	return 0;
}


CUTIL_API int xmlrpc_array_add_value (XmlrpcArray *xa, XmlrpcValue *value)
{
	int rc;
	if (xa == NULL || value == NULL)
		return EINVAL;
	rc = slist_item_add (&xa->valuelist, value);
	if (rc != 0)
		return rc;
	rc = xmlrpc_value_ref_inc (value);
	if (rc != 0)
		return rc;
	return 0;
}


CUTIL_API int xmlrpc_array_value_count (XmlrpcArray *xa, int *valuecount)
{
	int rc;
	if (xa == NULL || valuecount == NULL)
		return EINVAL;
	*valuecount = 0;
	if (xa->valuelist == NULL)
		return 0;
	rc = slist_item_count (xa->valuelist, valuecount);
	return rc;
}

CUTIL_API int xmlrpc_array_get_value (XmlrpcArray *xa, int valuenumber,
				XmlrpcValue **xvp)
{
	int rc;
	if (xa == NULL || xvp == NULL)
		return EINVAL;
	*xvp = NULL;
	rc = slist_item_get (xa->valuelist, valuenumber,
				(void **)xvp);
	if (rc == 0)
		xmlrpc_value_ref_inc (*xvp);
	return rc;
}


/************************************************************************/
/* XMLRPC request functions */
/************************************************************************/
CUTIL_API int xmlrpc_request_new (XmlrpcRequest **xrp, char *methodname)
{
	XmlrpcRequest *xr = NULL;
	if (xrp == NULL || methodname == NULL)
		return EINVAL;
	*xrp = NULL;
	xr = (XmlrpcRequest *)calloc (1, sizeof (XmlrpcRequest));
	if (xr == NULL)
		return ENOMEM;
	xr->methodname = strdup (methodname);
	if (xr->methodname == NULL) {
		free (xr);
		return ENOMEM;
	}
	*xrp = xr;
	return 0;
}


CUTIL_API int xmlrpc_request_add_param (XmlrpcRequest *xr, XmlrpcValue *param)
{
	int rc;
	if (xr == NULL || param == NULL)
		return EINVAL;
	rc = xmlrpc_value_ref_inc (param);
	rc = slist_item_add (&xr->paramlist, param);
	return rc;
}


CUTIL_API int xmlrpc_request_get_param_count (XmlrpcRequest *xr, int *paramcount)
{
	int rc;
	if (xr == NULL || paramcount == NULL)
		return EINVAL;
	*paramcount = 0;
	if (xr->paramlist == NULL)
		return 0;
	rc = slist_item_count (xr->paramlist, paramcount);
	return rc;
}


CUTIL_API int xmlrpc_request_get_param (XmlrpcRequest *xr, int paramnumber,
			XmlrpcValue **xvp)
{
	int rc;
	if (xr == NULL || paramnumber < 0 || xvp == NULL)
		return EINVAL;
	*xvp = NULL;
	if (xr->paramlist == NULL)
		return ENOENT;
	rc = slist_item_get (xr->paramlist, paramnumber, (void **)xvp);
	if (rc == 0)
		xmlrpc_value_ref_inc (*xvp);
	return rc;
}


CUTIL_API int xmlrpc_request_get_methodname (XmlrpcRequest *xr, const char **methodnamep)
{
	if (xr == NULL || methodnamep == NULL)
		return EINVAL;
	*methodnamep = xr->methodname;
	return 0;
}


CUTIL_API int xmlrpc_request_serialize (XmlrpcRequest *xr, char **outputxmlp,
			int *outputxmllen)
{
	int rc;
	SList *litem;
	XmlrpcValue *xv = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr rootnode = NULL;
	xmlNodePtr paramsnode = NULL;
	xmlNodePtr pnode = NULL;
	xmlNodePtr vnode = NULL;
	
	if (xr == NULL || outputxmlp == NULL || outputxmllen == NULL)
		return EINVAL;
	*outputxmlp = NULL;
	*outputxmllen = 0;

	doc = xmlNewDoc (BAD_CAST "1.0");
	if (doc == NULL)
		return ENOMEM;
	rootnode = xmlNewNode (NULL, BAD_CAST "methodCall");
	if (rootnode == NULL) {
		xmlFreeDoc (doc);
		return ENOMEM;
	}
	xmlDocSetRootElement (doc, rootnode);

	xmlNewChild (rootnode,
			NULL,
			BAD_CAST "methodName",
			(unsigned char*)xr->methodname);
	paramsnode = xmlNewChild (rootnode, NULL, BAD_CAST "params", NULL);
	if (paramsnode == NULL) {
		xmlFreeDoc (doc);
		return ENOMEM;
	}

	litem = xr->paramlist;
	while (litem != NULL) {
		xv = slist_get_data (litem);
		pnode = xmlNewChild (paramsnode,
					NULL,
					BAD_CAST "param",
					NULL);
		rc = xmlrpc_value_serialize_xml (&vnode, xv);
		xmlAddChild (pnode, vnode);
		litem = slist_get_next (litem);
	}
	
	xmlDocDumpMemory (doc, (xmlChar **)outputxmlp, outputxmllen);
	xmlFreeDoc (doc);
	if (*outputxmlp == NULL)
		return ENOMEM;
	return 0;
}


CUTIL_API int xmlrpc_request_parse_xml (XmlrpcRequest **xrp, char *inputxml,
			int inputxmllen)
{
	int rc;
	char *value = NULL;
	XmlrpcRequest *xr = NULL;
	XmlrpcValue *xv = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr mcnode = NULL;
	xmlNodePtr mnnode = NULL;
	xmlNodePtr psnode = NULL;
	xmlNodePtr pnode = NULL;
	xmlNodePtr vnode = NULL;
	if (xrp == NULL || inputxml == NULL || inputxmllen < 1)
		return EINVAL;
	*xrp = NULL;
	doc = xmlParseMemory (inputxml, inputxmllen);
	if (doc == NULL)
		return ENOMEM;
	mcnode = xmlDocGetRootElement (doc);
	if (mcnode == NULL) {
		xmlFreeDoc (doc);
		return ENOMEM;
	}
	if (strcmp ((char*)mcnode->name, "methodCall")) {
		xmlFreeDoc (doc);
		return EINVAL;
	}
	mnnode = mcnode->children;
	if (mnnode == NULL || strcmp ((char*)mnnode->name, "methodName")) {
		xmlFreeDoc (doc);
		return EINVAL;
	}
	psnode = mnnode->next;
	if (psnode == NULL || strcmp ((char*)psnode->name, "params")) {
		xmlFreeDoc (doc);
		return EINVAL;
	}
	value = (char*)xmlNodeGetContent(mnnode);
	rc = xmlrpc_request_new (&xr, value);
	if (rc != 0) {
		xmlFree (value);
		xmlFreeDoc (doc);
		return EINVAL;
	}
	xmlFree (value);
	for (pnode = psnode->children ; pnode != NULL ; pnode = pnode->next) {
		if (!strcmp ((char*)pnode->name, "param")) {
			vnode = pnode->children;
			if ( (vnode != NULL) && (!strcmp ((char*)vnode->name,
							"value"))) {
				rc = xmlrpc_value_new_parse_xml (&xv, vnode);
				if (rc != 0) {
					xmlrpc_request_free (&xr);
					xmlFreeDoc (doc);
					return rc;
				}
				rc = xmlrpc_request_add_param (xr, xv);
				if (rc != 0) {
					xmlrpc_value_ref_dec (&xv);
					xmlrpc_request_free (&xr);
					xmlFreeDoc (doc);
					return rc;
				}
				xmlrpc_value_ref_dec (&xv);
			}
		}
	}
	xmlFreeDoc (doc);
	*xrp = xr;
	return 0;
}


CUTIL_API int xmlrpc_request_free (XmlrpcRequest **xrp)
{
	int rc;
	SList *litem;
	XmlrpcRequest *xr = NULL;
	XmlrpcValue *xv = NULL;
	if (xrp == NULL || *xrp == NULL)
		return EINVAL;
	xr = *xrp;

	litem = xr->paramlist;
	while (litem) {
		xv = slist_get_data (litem);
		rc = xmlrpc_value_ref_dec (&xv);
		litem = slist_get_next (litem);
	}
	slist_free (&xr->paramlist);


	if (xr->methodname != NULL)
		free (xr->methodname);
	free (xr);
	*xrp = NULL;
	return 0;
}



/************************************************************************/
/* XMLRPC response functions */
/************************************************************************/
CUTIL_API int xmlrpc_response_new (XmlrpcResponse **xrp, char *methodname,
			XmlrpcValue *xv)
{
	int rc;
	XmlrpcResponse *xr = NULL;
	if (xrp == NULL || methodname == NULL || xv == NULL)
		return EINVAL;
	*xrp = NULL;
	xr = (XmlrpcResponse *)calloc (1, sizeof (XmlrpcResponse));
	if (xr == NULL)
		return ENOMEM;
	xr->methodname = strdup (methodname);
	if (xr->methodname == NULL) {
		free (xr);
		return ENOMEM;
	}
	rc = slist_item_add (&xr->paramlist, xv);
	if (rc != 0) {
		free (xr->methodname);
		free (xr);
		return rc;
	}
	rc = xmlrpc_value_ref_inc (xv);
	xr->type = XMLRPC_RESPONSETYPE_NORMAL;
	*xrp = xr;
	return 0;
}


CUTIL_API int xmlrpc_response_fault_new (XmlrpcResponse **xrp, char *methodname,
			int faultcode, char *faultstring)
{
	int rc;
	XmlrpcResponse *xr = NULL;
	XmlrpcStruct *xs = NULL;
	XmlrpcValue *xv = NULL;
	if (xrp == NULL || methodname == NULL || faultstring == NULL)
		return EINVAL;
	*xrp = NULL;
	rc = xmlrpc_struct_new (&xs);
	if (rc != 0)
		return rc;

	rc = xmlrpc_value_new (&xv,
				XMLRPC_VALUETYPE_INT,
				&faultcode, 
				sizeof (faultcode));
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		return rc;
	}
	rc = xmlrpc_struct_add_member (xs, "faultCode", xv);
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		xmlrpc_value_ref_dec (&xv);
		return rc;
	}
	rc = xmlrpc_value_ref_dec (&xv);

	rc = xmlrpc_value_new (&xv,
				XMLRPC_VALUETYPE_STRING,
				faultstring, 
				strlen (faultstring));
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		return rc;
	}
	rc = xmlrpc_struct_add_member (xs, "faultString", xv);
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		xmlrpc_value_ref_dec (&xv);
		return rc;
	}
	rc = xmlrpc_value_ref_dec (&xv);

	rc = xmlrpc_value_new_from_struct (&xv, xs);
	if (rc != 0) {
		xmlrpc_struct_ref_dec (&xs);
		return rc;
	}
	rc = xmlrpc_struct_ref_dec (&xs);

	rc = xmlrpc_response_new (&xr, methodname, xv);
	if (rc != 0) {
		xmlrpc_value_ref_dec (&xv);
		return rc;
	}
	xr->type = XMLRPC_RESPONSETYPE_FAULT;
	rc = xmlrpc_value_ref_dec (&xv);
	*xrp = xr;
	return 0;
}


CUTIL_API int xmlrpc_response_fault_new_with_value (XmlrpcResponse **xrp,
			char *methodname,
			XmlrpcValue *xv)
{
	int rc;
	XmlrpcResponse *xr = NULL;
	if (xrp == NULL || methodname == NULL || xv == NULL)
		return EINVAL;
	*xrp = NULL;
	rc = xmlrpc_response_new (&xr, methodname, xv);
	if (rc != 0)
		return rc;
	xr->type = XMLRPC_RESPONSETYPE_FAULT;
	*xrp = xr;
	return 0;
}


CUTIL_API int xmlrpc_response_get_param_count (XmlrpcResponse *xr, int *paramcount)
{
	int rc;
	if (xr == NULL || paramcount == NULL)
		return EINVAL;
	*paramcount = 0;
	if (xr->paramlist == NULL)
		return 0;
	rc = slist_item_count (xr->paramlist, paramcount);
	return rc;
}


CUTIL_API int xmlrpc_response_get_param (XmlrpcResponse *xr, int paramnumber,
			XmlrpcValue **xvp)
{
	int rc;
	if (xr == NULL || paramnumber < 0 || xvp == NULL)
		return EINVAL;
	*xvp = NULL;
	if (xr->paramlist == NULL)
		return ENOENT;
	rc = slist_item_get (xr->paramlist, paramnumber, (void **)xvp);
	if (rc == 0)
		xmlrpc_value_ref_inc (*xvp);
	return rc;
}


CUTIL_API int xmlrpc_response_get_methodname (XmlrpcResponse *xr, const char **methodnamep)
{
	if (xr == NULL || methodnamep == NULL)
		return EINVAL;
	*methodnamep = xr->methodname;
	return 0;
}


CUTIL_API int xmlrpc_response_serialize (XmlrpcResponse *xr, char **outputxmlp,
			int *outputxmllen)
{
	int rc;
	SList *litem;
	XmlrpcValue *xv = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr rootnode = NULL;
	xmlNodePtr paramsnode = NULL;
	xmlNodePtr pnode = NULL;
	xmlNodePtr vnode = NULL;
	
	if (xr == NULL || outputxmlp == NULL || outputxmllen == NULL)
		return EINVAL;
	*outputxmlp = NULL;
	*outputxmllen = 0;

	doc = xmlNewDoc (BAD_CAST "1.0");
	if (doc == NULL)
		return ENOMEM;
	rootnode = xmlNewNode (NULL, BAD_CAST "methodResponse");
	if (rootnode == NULL) {
		xmlFreeDoc (doc);
		return ENOMEM;
	}
	xmlDocSetRootElement (doc, rootnode);

/*
 * The XMLRPC spec doesn't allow this, but in the future maybe it will!
 * (ie. for allowing/differentiating multiple concurrent async calls)
*/
/*
	xmlNewChild (rootnode, NULL, BAD_CAST "methodName", xr->methodname);
*/

	if (xr->type == XMLRPC_RESPONSETYPE_NORMAL) {
		paramsnode = xmlNewChild (rootnode, NULL, BAD_CAST "params", NULL);
		if (paramsnode == NULL) {
			xmlFreeDoc (doc);
			return ENOMEM;
		}
		litem = xr->paramlist;
		while (litem != NULL) {
			xv = slist_get_data (litem);
			pnode = xmlNewChild (paramsnode,
						NULL,
						BAD_CAST "param",
						NULL);
			rc = xmlrpc_value_serialize_xml (&vnode, xv);
			xmlAddChild (pnode, vnode);
			litem = slist_get_next (litem);
		}

	} else if (xr->type == XMLRPC_RESPONSETYPE_FAULT) {
		paramsnode = xmlNewChild (rootnode, NULL, BAD_CAST "fault", NULL);
		if (paramsnode == NULL) {
			xmlFreeDoc (doc);
			return ENOMEM;
		}

		/* loop through params (should only be ONE for a fault) */
		litem = xr->paramlist;
		while (litem != NULL) {
			xv = slist_get_data (litem);
			rc = xmlrpc_value_serialize_xml (&vnode,
						xv);
			xmlAddChild (paramsnode, vnode);
			litem = slist_get_next (litem);
		}
	}

	xmlDocDumpMemory (doc, (xmlChar **)outputxmlp, outputxmllen);
	xmlFreeDoc (doc);
	if (*outputxmlp == NULL)
		return ENOMEM;
	return 0;
}


CUTIL_API int xmlrpc_response_parse_xml (XmlrpcResponse **xrp, char *inputxml,
			int inputxmllen)
{
	int rc;
	int responsetype;
	XmlrpcResponse *xr = NULL;
	XmlrpcValue *xv = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr mrnode = NULL;
	xmlNodePtr psnode = NULL;
	xmlNodePtr pnode = NULL;
	xmlNodePtr vnode = NULL;
	if (xrp == NULL || inputxml == NULL || inputxmllen < 1)
		return EINVAL;
	*xrp = NULL;
	doc = xmlParseMemory (inputxml, inputxmllen);
	if (doc == NULL)
		return ENOMEM;
	mrnode = xmlDocGetRootElement (doc);
	if (mrnode == NULL || strcmp ((char*)mrnode->name, "methodResponse")) {
		xmlFreeDoc (doc);
		return EINVAL;
	}

/*
 * The current XMLRPC spec. only allows this.
 * (ie. no methodName and only ONE return 'value'.
*/
	psnode = mrnode->children;
	if (psnode == NULL) {
		xmlFreeDoc (doc);
		return EINVAL;
	}
	if (!strcmp ((char*)psnode->name, "params")) {
		pnode = psnode->children;
		if (pnode == NULL || strcmp ((char*)pnode->name, "param")) {
			xmlFreeDoc (doc);
			return EINVAL;
		}
		vnode = pnode->children;
		if (vnode == NULL || strcmp ((char*)vnode->name, "value")) {
			xmlFreeDoc (doc);
			return EINVAL;
		}
		responsetype = XMLRPC_RESPONSETYPE_NORMAL;
	} else if (!strcmp ((char*)psnode->name, "fault")) {
		vnode = psnode->children;
		if (vnode == NULL || strcmp ((char*)vnode->name, "value")) {
			xmlFreeDoc (doc);
			return EINVAL;
		}
		responsetype = XMLRPC_RESPONSETYPE_FAULT;
	} else {
		xmlFreeDoc (doc);
		return EINVAL;
	}
	rc = xmlrpc_value_new_parse_xml (&xv, vnode);
	if (rc != 0) {
		xmlFreeDoc (doc);
		return rc;
	}
	rc = xmlrpc_response_new (&xr, "", xv);
	if (rc != 0) {
		xmlFreeDoc (doc);
		xmlrpc_value_ref_dec (&xv);
		return rc;
	}
	xmlrpc_value_ref_dec (&xv);
	xr->type = responsetype;

	xmlFreeDoc (doc);
	*xrp = xr;
	return 0;
}



CUTIL_API int xmlrpc_response_free (XmlrpcResponse **xrp)
{
	int rc;
	SList *litem;
	XmlrpcResponse *xr = NULL;
	XmlrpcValue *xv = NULL;
	if (xrp == NULL || *xrp == NULL)
		return EINVAL;
	xr = *xrp;

	litem = xr->paramlist;
	while (litem != NULL) {
		xv = slist_get_data (litem);
		rc = xmlrpc_value_ref_dec (&xv);
		litem = slist_get_next (litem);
	}
	slist_free (&xr->paramlist);

	if (xr->methodname != NULL)
		free (xr->methodname);
	free (xr);
	*xrp = NULL;
	return 0;
}



/************************************************************************/
/* Internal functions */
/************************************************************************/
int xmlrpc_value_serialize_xml (xmlNodePtr *nodep, XmlrpcValue *xv)
{
	int rc;
	char *str = NULL;
	SList *litem = NULL;
	XmlrpcValue *xv1 = NULL;
	XmlrpcStruct *xs = NULL;
	XmlrpcStructMember *xsm = NULL;
	XmlrpcArray *xa = NULL;
	xmlNodePtr n = NULL;
	xmlNodePtr node = NULL;
	xmlNodePtr dnode = NULL;
	xmlNodePtr pnode = NULL;
	xmlNodePtr vnode = NULL;
	if (nodep == NULL || xv == NULL)
		return EINVAL;

	*nodep = NULL;

	str = (char *)calloc (1, XMLRPC_STRBUFFER_SIZE);
	if (str == NULL)
		return ENOMEM;

	vnode = xmlNewNode (NULL, BAD_CAST "value");
	if (vnode == NULL) {
		free (str);
		return ENOMEM;
	}

	switch (xv->type) {
		case XMLRPC_VALUETYPE_INT:
			rc = snprintf (str,
					XMLRPC_STRBUFFER_SIZE,
					"%ld", *(long *)xv->data);
			n = xmlNewChild (vnode,
					NULL,
					BAD_CAST "i4",
					BAD_CAST str);
			if (n == NULL) {
				xmlFreeNode (vnode);
				free (str);
				return ENOMEM;
			}
			break;

		case XMLRPC_VALUETYPE_BOOL:
			rc = snprintf (str,
					XMLRPC_STRBUFFER_SIZE,
					"%ld", *(long *)xv->data);
			n = xmlNewChild (vnode,
					NULL,
					BAD_CAST "boolean",
					BAD_CAST str);
			if (n == NULL) {
				xmlFreeNode (vnode);
				free (str);
				return ENOMEM;
			}
			break;

		case XMLRPC_VALUETYPE_STRING:
			n = xmlNewChild (vnode,
					NULL,
					BAD_CAST "string",
					BAD_CAST (char *)xv->data);
			if (n == NULL) {
				xmlFreeNode (vnode);
				free (str);
				return ENOMEM;
			}
			break;

		case XMLRPC_VALUETYPE_DOUBLE:
			rc = snprintf (str,
					XMLRPC_STRBUFFER_SIZE,
					"%f", *(double *)xv->data);
			n = xmlNewChild (vnode,
					NULL,
					BAD_CAST "double",
					BAD_CAST str);
			if (n == NULL) {
				xmlFreeNode (vnode);
				free (str);
				return ENOMEM;
			}
			break;

		case XMLRPC_VALUETYPE_DATETIME:
			break;

		case XMLRPC_VALUETYPE_BASE64:
			n = xmlNewChild (vnode,
					NULL,
					BAD_CAST "base64",
					BAD_CAST xv->data);
			if (n == NULL) {
				xmlFreeNode (vnode);
				free (str);
				return ENOMEM;
			}
			break;

		case XMLRPC_VALUETYPE_STRUCT:
			xs = (XmlrpcStruct *)xv->data;
			pnode = xmlNewNode (NULL, BAD_CAST "struct");
			litem = xs->memberlist;
			while (litem != NULL) {
				xsm = slist_get_data (litem);
				dnode = xmlNewNode (NULL,
						BAD_CAST "member");
				xmlNewChild (dnode,
						NULL,
						BAD_CAST "name",
						BAD_CAST xsm->name);
				rc = xmlrpc_value_serialize_xml (&node,
						xsm->value);
				xmlAddChild (dnode, node);
				xmlAddChild (pnode, dnode);
				litem = slist_get_next (litem);
			}
			xmlAddChild (vnode, pnode);
			break;

		case XMLRPC_VALUETYPE_ARRAY:
			xa = (XmlrpcArray *)xv->data;
			pnode = xmlNewNode (NULL, BAD_CAST "array");
			dnode = xmlNewNode (NULL, BAD_CAST "data");
			xmlAddChild (pnode, dnode);
			litem = xa->valuelist;
			while (litem != NULL) {
				xv1 = slist_get_data (litem);
				rc = xmlrpc_value_serialize_xml (&node,
						xv1);
				xmlAddChild (dnode, node);
				litem = slist_get_next (litem);
			}
			xmlAddChild (vnode, pnode);
			break;

		default:
			free (str);
			return EINVAL;
	}
	*nodep = vnode;
	free (str);
	return 0;
}


int xmlrpc_value_new_parse_xml (XmlrpcValue **xvp, xmlNodePtr valuenode)
{
	int rc;
	long lvalue;
	double dvalue;
	char *svalue = NULL;
	XmlrpcValue *xv = NULL;
	XmlrpcStruct *xs = NULL;
	XmlrpcArray *xa = NULL;
	xmlNodePtr n = NULL;
	if (xvp == NULL || valuenode == NULL)
		return EINVAL;
	*xvp = NULL;

	if (strcmp ((char*)valuenode->name, "value"))
		return EINVAL;

	n = valuenode->children;
	if ( (!strcmp ((char*)n->name, "i4")) || (!strcmp ((char*)n->name, "int")) ) {
		svalue = (char *)xmlNodeGetContent(n);
		lvalue = atol (svalue);
		rc = xmlrpc_value_new (&xv,
					XMLRPC_VALUETYPE_INT,
					&lvalue,
					sizeof (lvalue));
	} else if (!strcmp ((char*)n->name, "boolean")) {
		svalue = (char *)xmlNodeGetContent(n);
		if (svalue[0] == 't' || svalue[0] == 'T' || svalue[0] != '0')
			lvalue = 1;
		else
			lvalue = 0;
		rc = xmlrpc_value_new (&xv,
					XMLRPC_VALUETYPE_BOOL,
					&lvalue,
					sizeof (lvalue));
	} else if (!strcmp ((char*)n->name, "double")) {
		svalue = (char *)xmlNodeGetContent(n);
		dvalue = atof (svalue);
		rc = xmlrpc_value_new (&xv,
					XMLRPC_VALUETYPE_DOUBLE,
					&dvalue,
					sizeof (dvalue));
	} else if (!strcmp ((char*)n->name, "string")) {
		svalue = (char *)xmlNodeGetContent(n);
		rc = xmlrpc_value_new (&xv,
					XMLRPC_VALUETYPE_STRING,
					svalue,
					strlen (svalue));
	} else if (!strcmp ((char*)n->name, "datetime")) {
		svalue = (char *)xmlNodeGetContent(n);
		rc = xmlrpc_value_new (&xv,
					XMLRPC_VALUETYPE_DATETIME,
					svalue,
					strlen (svalue));
	} else if (!strcmp ((char*)n->name, "base64")) {
		svalue = (char *)xmlNodeGetContent(n);
		rc = xmlrpc_value_new (&xv,
					XMLRPC_VALUETYPE_BASE64,
					svalue,
					strlen (svalue));
	} else if (!strcmp ((char*)n->name, "struct")) {
		rc = xmlrpc_struct_new_parse_xml (&xs, n);
		if (rc != 0)
			return rc;
		rc = xmlrpc_value_new_from_struct (&xv, xs);
		if (rc != 0) {
			xmlrpc_struct_ref_dec (&xs);
			return rc;
		}
		xmlrpc_struct_ref_dec (&xs);
	} else if (!strcmp ((char*)n->name, "array")) {
		rc = xmlrpc_array_new_parse_xml (&xa, n);
		if (rc != 0)
			return rc;
		rc = xmlrpc_value_new_from_array (&xv, xa);
		if (rc != 0) {
			xmlrpc_array_ref_dec (&xa);
			return rc;
		}
		xmlrpc_array_ref_dec (&xa);
	} else {
		return EINVAL;
	}		
	*xvp = xv;
	if (svalue != NULL)
		xmlFree (svalue);
	return 0;
}



int xmlrpc_struct_new_parse_xml (XmlrpcStruct **xsp, xmlNodePtr structnode)
{
	int rc;
	char *value = NULL;
	XmlrpcValue *xv = NULL;
	XmlrpcStruct *xs = NULL;
	xmlNodePtr mnode = NULL;
	xmlNodePtr nnode = NULL;
	xmlNodePtr vnode = NULL;
	if (xsp == NULL || structnode == NULL)
		return EINVAL;
	*xsp = NULL;
	if (strcmp ((char*)structnode->name, "struct"))
		return EINVAL;
	rc = xmlrpc_struct_new (&xs);
	if (rc != 0)
		return rc;
	for (mnode = structnode->children ; mnode != NULL ; mnode = mnode->next) {
		if (strcmp ((char*)mnode->name, "member")) {
			xmlrpc_struct_ref_dec (&xs);
			return EINVAL;
		}
		nnode = mnode->children;
		if (nnode == NULL || strcmp ((char*)nnode->name, "name")) {
			xmlrpc_struct_ref_dec (&xs);
			return EINVAL;
		}
		vnode = nnode->next;
		if (vnode == NULL || strcmp ((char*)vnode->name, "value")) {
			xmlrpc_struct_ref_dec (&xs);
			return EINVAL;
		}
		rc = xmlrpc_value_new_parse_xml (&xv, vnode);
		if (rc != 0 ) {
			xmlrpc_struct_ref_dec (&xs);
			return EINVAL;
		}
		value = (char*)xmlNodeGetContent(nnode);
		rc = xmlrpc_struct_add_member (xs,
					value,
					xv);
		if (rc != 0 ) {
			xmlrpc_value_ref_dec (&xv);
			xmlrpc_struct_ref_dec (&xs);
			xmlFree (value);
			return EINVAL;
		}
		xmlFree (value);
		xmlrpc_value_ref_dec (&xv);
	}
	*xsp = xs;
	return 0;
}


int xmlrpc_array_new_parse_xml (XmlrpcArray **xap, xmlNodePtr arraynode)
{
	int rc;
	XmlrpcValue *xv = NULL;
	XmlrpcArray *xa = NULL;
	xmlNodePtr dnode = NULL;
	xmlNodePtr vnode = NULL;

	if (xap == NULL || arraynode == NULL)
		return EINVAL;
	*xap = NULL;
	if (strcmp ((char*)arraynode->name, "array"))
		return EINVAL;
	dnode = arraynode->children;
	if (dnode == NULL || strcmp ((char*)dnode->name, "data"))
		return EINVAL;
	rc = xmlrpc_array_new (&xa);
	if (rc != 0)
		return rc;
	for (vnode = dnode->children ; vnode != NULL ; vnode = vnode->next) {
		rc = xmlrpc_value_new_parse_xml (&xv, vnode);
		if (rc != 0 ) {
			xmlrpc_array_ref_dec (&xa);
			return EINVAL;
		}
		rc = xmlrpc_array_add_value (xa, xv);
		if (rc != 0 ) {
			xmlrpc_value_ref_dec (&xv);
			xmlrpc_array_ref_dec (&xa);
			return EINVAL;
		}
		xmlrpc_value_ref_dec (&xv);
	}
	*xap = xa;
	return 0;
}


